@@ -20,6 +20,7 @@
 import json
 import os
 import re
+import socket
 import subprocess
 import tempfile
 import urllib.error
@@ -35,6 +36,7 @@ from lintian_brush import (
     certainty_sufficient,
     certainty_to_confidence,
     version_string,
+    min_certainty,
     )
 from lintian_brush.vcs import (
     browse_url_from_repo_url,
@@ -42,12 +44,15 @@ from lintian_brush.vcs import (
     sanitize_url as sanitize_vcs_url,
     is_gitlab_site,
     split_vcs_url,
+    unsplit_vcs_url,
     )
 from urllib.request import urlopen, Request
 
+# Pecl is quite slow, so up the timeout a bit.
+PECL_URLLIB_TIMEOUT = 15
 ADDON_ONLY_FIELDS = ['Archive']
 KNOWN_HOSTING_SITES = [
-    'code.launchpad.net', 'github.com', 'launchpad.net']
+    'code.launchpad.net', 'github.com', 'launchpad.net', 'git.openstack.org']
 
 
 class UpstreamDatum(object):
@@ -95,6 +100,8 @@ class UpstreamDatum(object):
 # - X-SourceForge-Project
 # - X-Wiki
 # - X-Summary
+# - X-License
+# - X-Copyright
 
 # Supported, but unused.
 # - FAQ
@@ -153,6 +160,10 @@ def guess_repo_from_url(url, net_access=False):
     if parsed_url.netloc == 'download.gnome.org':
         if len(path_elements) >= 2 and path_elements[0] == 'sources':
             return 'https://gitlab.gnome.org/GNOME/%s.git' % path_elements[1]
+    if parsed_url.netloc == 'download.kde.org':
+        if len(path_elements) >= 2 and path_elements[0] in (
+                'stable', 'unstable'):
+            return 'https://anongit.kde.org/%s.git' % path_elements[1]
     if parsed_url.netloc == 'ftp.gnome.org':
         if (len(path_elements) >= 4 and [
               e.lower() for e in path_elements[:3]] == [
@@ -200,8 +211,13 @@ def guess_repo_from_url(url, net_access=False):
     if parsed_url.netloc in KNOWN_HOSTING_SITES:
         return url
     # Maybe it's already pointing at a VCS repo?
-    if net_access and probe_upstream_branch_url(url):
+    if parsed_url.netloc.startswith('svn.'):
+        # 'svn' subdomains are often used for hosting SVN repositories.
         return url
+    if net_access:
+        if verify_repository_url(url):
+            return url
+        return None
     return None
 
 
@@ -272,6 +288,18 @@ def get_python_pkg_info(path, trust_package=False):
         return read_python_pkg_info(td)
 
 
+def guess_from_debian_rules(path, trust_package):
+    from .rules import Makefile
+    mf = Makefile.from_path(path)
+    try:
+        upstream_git = mf.get_variable(b'UPSTREAM_GIT')
+    except KeyError:
+        pass
+    else:
+        yield UpstreamDatum(
+            "Repository", sanitize_vcs_url(upstream_git.decode()), "certain")
+
+
 def guess_from_debian_watch(path, trust_package):
     from lintian_brush.watch import (
         parse_watch_file,
@@ -453,16 +481,17 @@ def guess_from_debian_copyright(path, trust_package):
                 "Contact", ','.join(header.upstream_contact), 'certain')
         if header.source:
             if ' 'in header.source:
-                from_url = re.split('[ ,]', header.source)[0]
+                from_urls = [u for u in re.split('[ ,\n]', header.source) if u]
             else:
-                from_url = header.source
-            repo_url = guess_repo_from_url(from_url)
-            if repo_url:
-                yield UpstreamDatum(
-                    'Repository', sanitize_vcs_url(repo_url), 'likely')
-            if (from_url.startswith('https://pecl.php.net/package/') or
-                    from_url.startswith('http://pecl.php.net/package/')):
-                yield UpstreamDatum('X-Pecl-URL', from_url, 'certain')
+                from_urls = [header.source]
+            for from_url in from_urls:
+                repo_url = guess_repo_from_url(from_url)
+                if repo_url:
+                    yield UpstreamDatum(
+                        'Repository', sanitize_vcs_url(repo_url), 'likely')
+                if (from_url.startswith('https://pecl.php.net/package/') or
+                        from_url.startswith('http://pecl.php.net/package/')):
+                    yield UpstreamDatum('X-Pecl-URL', from_url, 'certain')
         if "X-Upstream-Bugs" in header:
             yield UpstreamDatum(
                 "Bug-Database", header["X-Upstream-Bugs"], 'certain')
@@ -480,8 +509,9 @@ def guess_from_readme(path, trust_package):
         with open(path, 'rb') as f:
             lines = list(f.readlines())
             for i, line in enumerate(lines):
-                if line.strip().startswith(b'git clone'):
-                    line = line.strip()
+                line = line.strip()
+                if line.strip().lstrip(b'$').strip().startswith(b'git clone'):
+                    line = line.strip().lstrip(b'$').strip()
                     while line.endswith(b'\\'):
                         line += lines[i+1]
                         line = line.strip()
@@ -516,6 +546,16 @@ def guess_from_readme(path, trust_package):
                 if m:
                     yield UpstreamDatum(
                         'Repository', line.strip().decode(), 'likely')
+                m = re.fullmatch(b'git://([^ ]+)', line)
+                if m:
+                    yield UpstreamDatum(
+                        'Repository', line.strip().decode(), 'likely')
+                for m in re.finditer(
+                        b'https://([^/]+)/([^ ]+)', line):
+                    if is_gitlab_site(m.group(1).decode()):
+                        yield UpstreamDatum(
+                            'Repository',
+                            m.group(0).decode().rstrip(), 'likely')
     except IsADirectoryError:
         pass
 
@@ -537,6 +577,10 @@ def guess_from_debian_patch(path, trust_package):
                 bug_db = bug_database_from_issue_url(forwarded.decode('utf-8'))
                 if bug_db:
                     yield UpstreamDatum('Bug-Database', bug_db, 'possible')
+                repo_url = repo_url_from_merge_request_url(
+                    forwarded.decode('utf-8'))
+                if repo_url:
+                    yield UpstreamDatum('Repository', repo_url, 'possible')
 
 
 def guess_from_meta_json(path, trust_package):
@@ -572,7 +616,7 @@ def guess_from_meta_yml(path, trust_package):
     """
     import ruamel.yaml
     import ruamel.yaml.reader
-    with open(path, 'r') as f:
+    with open(path, 'rb') as f:
         try:
             data = ruamel.yaml.load(f, ruamel.yaml.SafeLoader)
         except ruamel.yaml.reader.ReaderError as e:
@@ -659,6 +703,60 @@ def guess_from_doap(path, trust_package):
                             'Repository-Browse', web_url, 'certain')
 
 
+def guess_from_cabal(path, trust_package=False):
+    # TODO(jelmer): Perhaps use a standard cabal parser in Python?
+    # The current parser is not really correct, but good enough for our needs.
+    # https://www.haskell.org/cabal/release/cabal-1.10.1.0/doc/users-guide/
+    repo_url = None
+    repo_branch = None
+    repo_subpath = None
+    with open(path, 'r', encoding='utf-8') as f:
+        section = None
+        for line in f:
+            if line.lstrip().startswith('--'):
+                # Comment
+                continue
+            if not line.strip():
+                section = None
+                continue
+            try:
+                (field, value) = line.split(':', 1)
+            except ValueError:
+                if not line.startswith(' '):
+                    section = line.strip().lower()
+                continue
+            # The case of field names is not sigificant
+            field = field.lower()
+            value = value.strip()
+            if not field.startswith(' '):
+                if field == 'homepage':
+                    yield UpstreamDatum('Homepage', value, 'certain')
+                if field == 'bug-reports':
+                    yield UpstreamDatum('Bug-Database', value, 'certain')
+                if field == 'name':
+                    yield UpstreamDatum('Name', value, 'certain')
+                if field == 'maintainer':
+                    yield UpstreamDatum('Contact', value, 'certain')
+                if field == 'copyright':
+                    yield UpstreamDatum('X-Copyright', value, 'certain')
+                if field == 'license':
+                    yield UpstreamDatum('X-License', value, 'certain')
+            else:
+                field = field.strip()
+                if section == 'source-repository head':
+                    if field == 'location':
+                        repo_url = value
+                    if field == 'branch':
+                        repo_branch = value
+                    if field == 'subdir':
+                        repo_subpath = value
+    if repo_url:
+        yield UpstreamDatum(
+            'Repository',
+            unsplit_vcs_url(repo_url, repo_branch, repo_subpath),
+            'certain')
+
+
 def is_email_address(value):
     return '@' in value or ' (at) ' in value
 
@@ -756,6 +854,7 @@ def _get_guessers(path, trust_package=False):
     CANDIDATES = [
         ('debian/watch', guess_from_debian_watch),
         ('debian/control', guess_from_debian_control),
+        ('debian/rules', guess_from_debian_rules),
         ('setup.py', guess_from_setup_py),
         ('package.json', guess_from_package_json),
         ('package.xml', guess_from_package_xml),
@@ -778,10 +877,19 @@ def _get_guessers(path, trust_package=False):
             warn('More than one doap filename, ignoring all: %r' %
                  doap_filenames)
 
+    cabal_filenames = [n for n in os.listdir(path) if n.endswith('.cabal')]
+    if cabal_filenames:
+        if len(cabal_filenames) == 1:
+            CANDIDATES.append((cabal_filenames[0], guess_from_cabal))
+        else:
+            warn('More than one cabal filename, ignoring all: %r' %
+                 cabal_filenames)
+
     readme_filenames = [
         n for n in os.listdir(path)
-        if n.startswith('README') and
-        os.path.splitext(n)[1] not in ('.html', '.pdf', '.xml')]
+        if any([n.startswith(p)
+                for p in ['README', 'HACKING', 'CONTRIBUTING']])
+        and os.path.splitext(n)[1] not in ('.html', '.pdf', '.xml')]
     CANDIDATES.extend([(n, guess_from_readme) for n in readme_filenames])
 
     try:
@@ -865,8 +973,27 @@ def _possible_fields_missing(upstream_metadata, fields, field_certainty):
         return False
 
 
+def _sf_git_extract_url(page):
+    from bs4 import BeautifulSoup
+    bs = BeautifulSoup(page, features='lxml')
+    el = bs.find(id='access_url')
+    if not el:
+        return None
+    value = el.get('value')
+    if value is None:
+        return None
+    access_command = value.split(' ')
+    if access_command[:2] != ['git', 'clone']:
+        return None
+    return access_command[2]
+
+
 def guess_from_sf(sf_project):
-    data = get_sf_metadata(sf_project)
+    try:
+        data = get_sf_metadata(sf_project)
+    except socket.timeout:
+        warn('timeout contacting launchpad, ignoring: %s' % sf_project)
+        return
     if data.get('name'):
         yield 'Name', data['name']
     if data.get('external_homepage'):
@@ -885,14 +1012,20 @@ def guess_from_sf(sf_project):
     if len(vcs_tools) == 1:
         (kind, url) = vcs_tools[0]
         if kind == 'git':
-            url = urljoin('https://git.code.sf.net/', url)
+            url = urljoin('https://sourceforge.net/', url)
+            headers = {'User-Agent': USER_AGENT, 'Accept': 'text/html'}
+            http_contents = urlopen(
+                Request(url, headers=headers),
+                timeout=DEFAULT_URLLIB_TIMEOUT).read()
+            url = _sf_git_extract_url(http_contents)
         elif kind == 'svn':
             url = urljoin('https://svn.code.sf.net/', url)
         elif kind == 'hg':
             url = urljoin('https://hg.code.sf.net/', url)
         else:
             raise KeyError(kind)
-        yield 'Repository', url
+        if url is not None:
+            yield 'Repository', url
 
 
 def extend_from_external_guesser(
@@ -958,6 +1091,27 @@ def extract_sf_project_name(url):
     m = re.fullmatch('https?://(.*).(sf|sourceforge).net/?', url)
     if m:
         return m.group(1)
+    m = re.match('https://sourceforge.net/projects/([^/]+)', url)
+    if m:
+        return m.group(1)
+
+
+def repo_url_from_merge_request_url(url):
+    parsed_url = urlparse(url)
+    if parsed_url.netloc == 'github.com':
+        path_elements = parsed_url.path.strip('/').split('/')
+        if len(path_elements) > 2 and path_elements[2] == 'issues':
+            return urlunparse(
+                ('https', 'github.com', '/'.join(path_elements[:3]),
+                 None, None, None))
+    if is_gitlab_site(parsed_url.netloc):
+        path_elements = parsed_url.path.strip('/').split('/')
+        if (len(path_elements) > 2 and
+                path_elements[-2] == 'merge_requests' and
+                path_elements[-1].isdigit()):
+            return urlunparse(
+                ('https', parsed_url.netloc, '/'.join(path_elements[:-2]),
+                 None, None, None))
 
 
 def bug_database_from_issue_url(url):
@@ -968,6 +1122,14 @@ def bug_database_from_issue_url(url):
             return urlunparse(
                 ('https', 'github.com', '/'.join(path_elements[:3]),
                  None, None, None))
+    if is_gitlab_site(parsed_url.netloc):
+        path_elements = parsed_url.path.strip('/').split('/')
+        if (len(path_elements) > 2 and
+                path_elements[-2] == 'issues' and
+                path_elements[-1].isdigit()):
+            return urlunparse(
+                ('https', parsed_url.netloc, '/'.join(path_elements[:-2]),
+                 None, None, None))
 
 
 def guess_bug_database_url_from_repo_url(url):
@@ -992,6 +1154,33 @@ def guess_bug_database_url_from_repo_url(url):
     return None
 
 
+def bug_database_url_from_bug_submit_url(url):
+    parsed_url = urlparse(url)
+    path_elements = parsed_url.path.strip('/').split('/')
+    if parsed_url.netloc == 'github.com':
+        if len(path_elements) not in (3, 4):
+            return None
+        if path_elements[2] != 'issues':
+            return None
+        return urlunparse(
+            ('https', 'github.com', '/'.join(path_elements[:3]),
+             None, None, None))
+    if parsed_url.netloc == 'bugs.launchpad.net':
+        if len(path_elements) >= 1:
+            return urlunparse(
+                parsed_url._replace(path='/%s' % path_elements[0]))
+    if is_gitlab_site(parsed_url.netloc):
+        if len(path_elements) < 2:
+            return None
+        if path_elements[-2] != 'issues':
+            return None
+        if path_elements[-1] == 'new':
+            path_elements.pop(-1)
+        return urlunparse(
+            parsed_url._replace(path='/'.join(path_elements)))
+    return None
+
+
 def bug_submit_url_from_bug_database_url(url):
     parsed_url = urlparse(url)
     path_elements = parsed_url.path.strip('/').split('/')
@@ -1017,6 +1206,45 @@ def bug_submit_url_from_bug_database_url(url):
     return None
 
 
+def verify_repository_url(url, version=None):
+    parsed_url = urlparse(url)
+    if parsed_url.netloc == 'github.com':
+        path_elements = parsed_url.path.strip('/').split('/')
+        if path_elements[1].endswith('.git'):
+            path_elements[1] = path_elements[1][:-4]
+        if len(path_elements) < 2:
+            return False
+        api_url = 'https://api.github.com/repos/%s/%s' % (
+            path_elements[0], path_elements[1])
+        try:
+            data = _load_json_url(api_url)
+        except urllib.error.HTTPError as e:
+            if e.status == 404:
+                return False
+            elif e.status == 403:
+                # Probably rate-limited. Let's just hope for the best.
+                pass
+            else:
+                raise
+        else:
+            if data.get('archived', False):
+                return False
+            if data['description'].startswith('Moved to '):
+                return False
+            if 'has moved' in data['description']:
+                return False
+            if data['description'].startswith('Mirror of '):
+                return False
+            homepage = data.get('homepage')
+            if homepage and is_gitlab_site(homepage):
+                return False
+            # TODO(jelmer): Look at the contents of the repository; if it
+            # contains just a single README file with < 10 lines, assume
+            # the worst.
+            # return data['clone_url']
+    return probe_upstream_branch_url(url, version=version)
+
+
 def verify_bug_database_url(url):
     parsed_url = urlparse(url)
     if parsed_url.netloc == 'github.com':
@@ -1031,7 +1259,7 @@ def verify_bug_database_url(url):
             if e.status == 404:
                 return False
             raise
-        return data['has_issues']
+        return data['has_issues'] and not data.get('archived', False)
     if is_gitlab_site(parsed_url.netloc):
         path_elements = parsed_url.path.strip('/').split('/')
         if len(path_elements) < 3 or path_elements[-1] != 'issues':
@@ -1061,21 +1289,29 @@ def _extrapolate_repository_from_homepage(upstream_metadata, net_access):
     repo = guess_repo_from_url(
             upstream_metadata['Homepage'].value, net_access=net_access)
     if repo:
-        return UpstreamDatum('Repository', repo, 'likely')
+        return UpstreamDatum(
+            'Repository', repo,
+            min_certainty(['likely', upstream_metadata['Homepage'].certainty]))
 
 
 def _extrapolate_repository_from_download(upstream_metadata, net_access):
     repo = guess_repo_from_url(
             upstream_metadata['X-Download'].value, net_access=net_access)
     if repo:
-        return UpstreamDatum('Repository', repo, 'likely')
+        return UpstreamDatum(
+            'Repository', repo,
+            min_certainty(
+                ['likely', upstream_metadata['X-Download'].certainty]))
 
 
 def _extrapolate_repository_from_bug_db(upstream_metadata, net_access):
     repo = guess_repo_from_url(
         upstream_metadata['Bug-Database'].value, net_access=net_access)
     if repo:
-        return UpstreamDatum('Repository', repo, 'likely')
+        return UpstreamDatum(
+            'Repository', repo,
+            min_certainty(
+                ['likely', upstream_metadata['Bug-Database'].certainty]))
 
 
 def _extrapolate_repository_browse_from_repository(
@@ -1104,7 +1340,10 @@ def _extrapolate_bug_database_from_repository(
     bug_db_url = guess_bug_database_url_from_repo_url(
         upstream_metadata['Repository'].value)
     if bug_db_url:
-        return UpstreamDatum('Bug-Database', bug_db_url, 'likely')
+        return UpstreamDatum(
+            'Bug-Database', bug_db_url,
+            min_certainty(
+                ['likely', upstream_metadata['Repository'].certainty]))
 
 
 def _extrapolate_bug_submit_from_bug_db(
@@ -1117,6 +1356,16 @@ def _extrapolate_bug_submit_from_bug_db(
             upstream_metadata['Bug-Database'].certainty)
 
 
+def _extrapolate_bug_db_from_bug_submit(
+        upstream_metadata, net_access):
+    bug_db_url = bug_database_url_from_bug_submit_url(
+        upstream_metadata['Bug-Submit'].value)
+    if bug_db_url:
+        return UpstreamDatum(
+            'Bug-Database', bug_db_url,
+            upstream_metadata['Bug-Submit'].certainty)
+
+
 def _copy_bug_db_field(upstream_metadata, net_access):
     ret = UpstreamDatum(
         'Bug-Database',
@@ -1138,6 +1387,7 @@ EXTRAPOLATE_FNS = [
     ('Repository', 'Bug-Database',
      _extrapolate_bug_database_from_repository),
     ('Bug-Database', 'Bug-Submit', _extrapolate_bug_submit_from_bug_db),
+    ('Bug-Submit', 'Bug-Database', _extrapolate_bug_db_from_bug_submit),
     ('X-Download', 'Repository', _extrapolate_repository_from_download),
 ]
 
@@ -1173,7 +1423,7 @@ def extend_upstream_metadata(upstream_metadata, path, minimum_certainty=None,
             pass
         else:
             extend_from_lp(upstream_metadata, minimum_certainty, package)
-        extend_from_aur(upstream_metadata, minimum_certainty, package)
+            extend_from_aur(upstream_metadata, minimum_certainty, package)
     pecl_url = upstream_metadata.get('X-Pecl-URL')
     if net_access and pecl_url:
         extend_from_pecl(upstream_metadata, pecl_url.value, pecl_url.certainty)
@@ -1202,6 +1452,25 @@ def extend_upstream_metadata(upstream_metadata, path, minimum_certainty=None,
                 changed = True
 
 
+def _version_in_tags(version, tag_names):
+    if version in tag_names:
+        return True
+    if 'v%s' % version in tag_names:
+        return True
+    if 'release/%s' % version in tag_names:
+        return True
+    if version.replace('.', '_') in tag_names:
+        return True
+    for tag_name in tag_names:
+        if tag_name.endswith('_' + version):
+            return True
+        if tag_name.endswith('-' + version):
+            return True
+        if tag_name.endswith('_%s' % version.replace('.', '_')):
+            return True
+    return False
+
+
 def probe_upstream_branch_url(url, version=None):
     parsed = urlparse(url)
     if parsed.scheme in ('git+ssh', 'ssh', 'bzr+ssh'):
@@ -1220,21 +1489,8 @@ def probe_upstream_branch_url(url, version=None):
             if not tag_names:
                 # Uhm, hmm
                 return True
-            if version in tag_names:
+            if _version_in_tags(version, tag_names):
                 return True
-            if 'v%s' % version in tag_names:
-                return True
-            if 'release/%s' % version in tag_names:
-                return True
-            if version.replace('.', '_') in tag_names:
-                return True
-            for tag_name in tag_names:
-                if tag_name.endswith('_' + version):
-                    return True
-                if tag_name.endswith('-' + version):
-                    return True
-                if tag_name.endswith('_%s' % version.replace('.', '_')):
-                    return True
             return False
         else:
             return True
@@ -1270,7 +1526,7 @@ def check_upstream_metadata(upstream_metadata, version=None):
     """
     repository = upstream_metadata.get('Repository')
     if repository and repository.certainty == 'likely':
-        if probe_upstream_branch_url(repository.value, version=version):
+        if verify_repository_url(repository.value, version=version):
             repository.certainty = 'certain'
             derived_browse_url = browse_url_from_repo_url(repository.value)
             browse_repo = upstream_metadata.get('Repository-Browse')
@@ -1339,7 +1595,8 @@ def guess_from_pecl(package):
     url = 'https://pecl.php.net/packages/%s' % php_package.replace('-', '_')
     data = dict(guess_from_pecl_url(url))
     try:
-        data['Repository'] = guess_repo_from_url(data['Repository-Browse'])
+        data['Repository'] = guess_repo_from_url(
+                data['Repository-Browse'], net_access=True)
     except KeyError:
         pass
     return data.items()
@@ -1350,11 +1607,14 @@ def guess_from_pecl_url(url):
     try:
         f = urlopen(
             Request(url, headers=headers),
-            timeout=DEFAULT_URLLIB_TIMEOUT)
+            timeout=PECL_URLLIB_TIMEOUT)
     except urllib.error.HTTPError as e:
         if e.status != 404:
             raise
         return
+    except socket.timeout:
+        warn('timeout contacting pecl, ignoring: %s' % url)
+        return
     from bs4 import BeautifulSoup
     bs = BeautifulSoup(f.read(), features='lxml')
     tag = bs.find('a', text='Browse Source')
@@ -1435,6 +1695,9 @@ def guess_from_launchpad(package, distribution=None, suite=None):
         if e.status != 404:
             raise
         return
+    except socket.timeout:
+        warn('timeout contacting launchpad, ignoring')
+        return
 
     productseries_url = sourcepackage_data.get('productseries_link')
     if not productseries_url:
